import 'dart:async';
import 'dart:io';
import 'crypto.dart';
import 'packet.dart';
import 'net_addr.dart';
import 'peer_info.dart';
import 'tail_io.dart';
import 'utils.dart';
import 'verival.dart';

class Node {
  final List<int> sk;
  final Tail tail;
  final File? cache;
  bool _closed = false;

  Node(this.sk, this.tail, [this.cache]) {
    listen();
  }

  static Future<Node> create(List<int> sk, NetAddr tail, [File? cache]) async {
    return Node(await Crypto.genSK(), await Tail.connect(tail), cache);
  }

  void close() {
    if (_closed) {
      return;
    }
    tail.close();
    _closed = true;
  }

  bool get closed {
    return _closed;
  }

  Future<void> _setLastPacket(List<int> packet) async {
    _lastPacket = packet;
    if (cache != null) {
      await cache?.writeAsBytes(packet);
    }
  }

  List<int>? _lastPacket;

  Future<List<int>?> _getLastPacket() async {
    if (cache == null) {
      return _lastPacket;
    }
    try {
      return await cache!.readAsBytes();
    } catch (error) {
      return null;
    }
  }

  Stream<Packet> get stream {
    return tail.stream.asyncMap((event) async {
      try {
        return Packet.decode(event);
      } catch (_) {
        return null;
      }
    }).where((event) {
      return event != null && Utils.equal(event.to, Crypto.derivePK(sk));
    }).asyncMap((event) async {
      await _setLastPacket(event!.encode());
      _delay = Duration(milliseconds: 50);
      return Packet(
        event.to,
        event.id,
        event.from,
        await Crypto.decrypt(
            await Crypto.deriveKey(sk, event.from), event.data),
      );
    });
  }

  Duration _delay = Duration(milliseconds: 50);

  void listen() {
    Timer.run(() async {
      for (; !_closed;) {
        try {
          await Future.delayed(_delay);
          _delay = Duration(milliseconds: 500);
          var last = await _getLastPacket();
          last ??=
              Packet(Crypto.derivePK(sk), 0, List.filled(64, 0), [0]).encode();
          tail.check(last);
        } catch (_) {}
      }
    });
  }

  Future<void> publish(String name) {
    return Verival.publish(PeerInfo(tail: tail.host, name: name), sk);
  }

  Future<void> send(List<int> to, List<int> data, [int id = 0]) async {
    if (id == 0) {
      id = genNowMsgID();
    }
    var packet = Packet(to, id, Crypto.derivePK(sk),
        await Crypto.encrypt(await Crypto.deriveKey(sk, to), data));
    var encoded = packet.encode();
    if (encoded.length > 400) {
      throw Exception("too large packet");
    }
    var peerInfo = await Verival.lookup(to);
    await Tail.push(peerInfo.tail, encoded);
  }
}
